// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.

/*
 * ArfManip.cc
 *
 *  Created on: Mar 5, 2013
 *      Author: aibarra
 */

#include "RspManip.h"
#include "Efluxer.h"
#include "param.h"

#include "RmfDataServer.h"
#include "MatrixColumn.h"
#include "VarMatrixColumn.h"
#include "ResponseRow.h"

RspManip::~RspManip()
{
	for (unsigned int i = 0 ; i < _rspNames.size() ; ++i)
		dataSetServer->remove(_rspNames[i]);
}


void RspManip::addResponses(std::vector<std::string> arf,
		std::vector<std::string> rmf, EpicInstrInfo *e)

{
	std::vector<ARF*> arfs;
	for (unsigned int i = 0 ; i< arf.size() ; ++i)
	{
		ARF *a = new ARF(arf[i]);
		arfs.push_back(a);
	}

	std::vector<RMF*> rmfs;

	for (unsigned int i = 0 ; i< rmf.size() ; ++i)
	{
		RMF *r = new RMF(rmf[i]);
		rmfs.push_back(r);
	}


	int tlmin = (e->getSpectra("SRC"))[0]->getTLMin();
	int tlmax = (e->getSpectra("SRC"))[0]->getTLMax();

	//Get the number of spectryum channels
	int npi = (e->getSpectra("SRC"))[0]->getNPI();

	for (unsigned int i = 0 ; i < arf.size() ; ++i)
	{
		stringstream outFileName;
		outFileName <<"TEST_rsp_";
		 combineRMFandARF(rmfs[i],arfs[i],npi);


		outFileName << i << ".ds";
		_rspNames.push_back(outFileName.str());
		writeRsp(rmfs[i],tlmin,tlmax,outFileName.str());
	}

	combineRSP(rmfs,e);
	writeRsp(rmfs[0],tlmin,tlmax,stringParameter("filersp"));

}



void RspManip::writeRsp(RMF *r, int tlmin, int tlmax,string outFName)
{
	 unsigned long _max_ngroups=0;
	int _max_nresponse_elements=0;
	errstr << "writing new RSP file " << Sas::message(Msg::App,Msg::Verbose);
	for (unsigned e=0; e<r->getRSP()->matrix.size(); ++e) { // Loop over the energies

		vector<double> response = r->getRSP()->matrix[e];
		double _threshold = 1e-7;
		ResponseRow responserow(response, _threshold);
		//ResponseRow responserow(r->getRSP()->matrix[e], _threshold);

		int ngroups = responserow.ngroups();

		if (ngroups > _max_ngroups)
		{
			_max_ngroups = ngroups;
		}

		// Determine row with largest number of elements above threshold

		int nresponse_elements = responserow.nresponse_elements();

		if (nresponse_elements > _max_nresponse_elements)
		{
			_max_nresponse_elements = nresponse_elements;
		}
	}


	DataSet *rmfset = 0;
	rmfset =  dataSetServer->clone(r->getFileName(),outFName, Dal::Modify);

	if(rmfset->hasAttribute("INSTRUME"))
		*rmfset->attribute("INSTRUME")="EPIC";
	*(rmfset->table("EBOUNDS")->attribute("INSTRUME")) = "EPIC";

	rmfset->deleteBlock("MATRIX");


	 Table* rmftable = rmfset->addTable( "MATRIX",
	 				   r->getRSP()->matrix.size(),"",1);

	 /* Create ENERG_LO and ENERG_HI columns and set up references to them */

	 Column* ebinsloCol = rmftable->addColumn("ENERG_LO", Column::Real32, "No comment", "keV");

	 ColumnData *_ebinsloColData = ebinsloCol->data();
	 real32 *elod = _ebinsloColData->real32Data();

	 Column* ebinshiCol = rmftable->addColumn("ENERG_HI", Column::Real32, "No comment", "keV");

	 ColumnData *_ebinshiColData = ebinshiCol->data();
	 real32 *ehid = _ebinshiColData->real32Data();

	 vector<double> elo = r->getRSP()->elo;
	 vector<double> ehi = r->getRSP()->ehi;

	 for  (unsigned int i = 0 ; i < elo.size() ; ++i )
	 {
		 elod[i] = elo[i];
		 ehid[i] = ehi[i];
	 }

	 Column* n_grpCol  = rmftable->addColumn("N_GRP",
			 	 	 	 	 	 Column::Int16,
			 	 	 	 	 	 "No comment", "");
	 ColumnData *_n_grpColData = n_grpCol->data();

	 vector<unsigned long> sizeVector_fchan;


	 sizeVector_fchan.push_back(_max_ngroups);
	 Column* f_chanCol = rmftable->addColumn("F_CHAN",
			 	 	 	 	 	 Column::Int16,
			 	 	 	 	 	 	 "No comment",
			 	 	 	 	 	 	 "",
			 	 	 	 	 	 	 sizeVector_fchan);
	 ColumnData *_f_chanColData = f_chanCol->data();
	 *(f_chanCol->addAttribute("TLMIN")) = tlmin;
	 *(f_chanCol->addAttribute("TLMAX")) = tlmax;

    vector<unsigned long> sizeVector_nchan;

     sizeVector_nchan.push_back(_max_ngroups);

     Column* n_chanCol = rmftable->addColumn("N_CHAN",
    		 	 	 	 	 	 Column::Int16,
    		 	 	 	 	 	 "No comment",
    		 	 	 	 	 	 "",
    		 	 	 	 	 	 sizeVector_nchan);

     ColumnData *_n_chanColData = n_chanCol->data();

     vector<unsigned long> sizeVector_matrix;

      sizeVector_matrix.push_back(_max_nresponse_elements);

      MatrixColumn *_matrixColumn =  new VarMatrixColumn(rmftable);


     for (unsigned e=0; e<r->getRSP()->matrix.size(); ++e) { // Loop over the energies

    		  vector<double> response = r->getRSP()->matrix[e];
    		 double _threshold = 1e-7;
    		 ResponseRow responserow(response, _threshold);

    	     	 responserow.writeTo(e,
    				 _matrixColumn,
    				 _n_grpColData,
    				 _f_chanColData,
    				 _n_chanColData,
    				 tlmin);
     }


     int detchans = r->getRSP()->matrix[0].size();
     addAttributesTo(rmftable,"RSP_MATRIX", detchans);

     if (rmftable->hasAttribute("INSTRUME"))
    		 *rmftable->attribute("INSTRUME") = "EPIC";

     delete _ebinsloColData;
     delete _ebinshiColData;
     delete _n_grpColData;
     delete _f_chanColData;
     delete _n_chanColData;
     delete _matrixColumn;


     dataSetServer->close(rmfset);


}


void RspManip::combineRMFandARF(RMF *r, ARF *a, int npi)
{
	//Spectrum table rows
	//unsigned npi=spectrumtable->rows();


	vector<float> ener(a->size());

	vector<double> ehi = r->getRSP()->ehi;
	vector<double> elo = r->getRSP()->elo;
    for (int i=0; i<a->size(); ++i)
    	ener[i] = (ehi[i] + elo[i]) / 2.0;


    vector<float> energy = a->energy();

  //  for (unsigned i=0 ; i<energy.size() ; ++i)
   // 	cout << "energy " << ener[i] << " ARF VALUE " << a->value(ener[i]) << endl;


    errstr << "multiplying matrices... " << Sas::message(Msg::App,Msg::Verbose);
    unsigned nenerg = a->size();  // No. of energies in ARF and RMF

    for (unsigned e=0; e<nenerg; ++e) { // Loop over the energies
    	float val = a->value(e);

    	for (unsigned i=0; i<npi; ++i) { // Loop over the PIs

    		//cout << "mat[" << e<<"]["<<i<<"] " << r->getRSP()->matrix[e][i]
    		//         << " ARF[ " << e << "] "<<a->value(ener[e]);

    		r->getRSP()->matrix[e][i] = r->getRSP()->matrix[e][i] * val;
    		//cout << " NEW mat " << r->getRSP()->matrix[e][i] << endl;
    	}

    }


}


void RspManip::combineRSP(vector<RMF*> rmfs,EpicInstrInfo *e)
{

	//Get the exposure time for each of the MOS spectra.
	string fileName = rmfs[0]->getFileName();
	string type  = "SRC";

	e->getExposure(0, type);

	//Calculate weights
	//According to the FTOOLS help, the sum of the weigths has to be 1.0
	//BEGIN {print (2.0/(1.0+(e2/e1)))}'`

	vector<double> exp;
	double totalExp=0.;
	for (unsigned n = 0 ; n < rmfs.size() ; ++n)
	{
		exp.push_back( e->getExposure(n, type));
		totalExp = totalExp + e->getExposure(n, type);
	}


	//addrmfs
	 unsigned npi = rmfs[0]->getRSP()->matrix[0].size();
	 unsigned nenerg = rmfs[0]->getRSP()->matrix.size();

	 for (unsigned e=0; e<nenerg; ++e) { // Loop over the energies
		for (unsigned i=0; i<npi; ++i) { // Loop over the PIs
			rmfs[0]->getRSP()->matrix[e][i] =
							(rmfs[0]->getRSP()->matrix[e][i]*exp[0]);
		}
	}


	for (unsigned n = 1 ; n < rmfs.size() ; ++n)
	{
		for (unsigned e=0; e<nenerg; ++e) { // Loop over the energies
			for (unsigned i=0; i<npi; ++i) { // Loop over the PIs

				rmfs[0]->getRSP()->matrix[e][i] =
						( (rmfs[0]->getRSP()->matrix[e][i] +
						rmfs[n]->getRSP()->matrix[e][i]*exp[n])) ;
			}
		}
	}

	for (unsigned e=0; e<nenerg; ++e) { // Loop over the energies
		for (unsigned i=0; i<npi; ++i) { // Loop over the PIs
			rmfs[0]->getRSP()->matrix[e][i] =
				 (rmfs[0]->getRSP()->matrix[e][i])* (double(exp.size())/totalExp);
		}
	}


}



void RspManip::addAttributesTo(Table* table,
                            const string& hduclas2, int detchans)
{
 PROC;

     *table->addAttribute("TELESCOP",    "mission/satellite name")                  = "XMM";

     *table->addAttribute("INSTRUME",    "instrument/detector name")                = "EMOS";

     *table->addAttribute("FILTER",      "filter in use")                           = "UNDEFINED";

     *table->addAttribute("RMFVERSN",    "OGIP classification of FITS format")      = "1992a";

     *table->addAttribute("CHANTYPE",    "Detector Channel Type in use (PHA or PI") = "PI";

     *table->addAttribute("DETCHANS",    "total number of detector channels")       = detchans;

     *table->addAttribute("HDUCLASS",    "Format conforms to OGIP standard")        = "OGIP";

     *table->addAttribute("HDUCLAS1",    "dataset relates to spectral response")    = "RESPONSE";

     *table->addAttribute("HDUCLAS2",    "dataset is a spectral response matrix")   = hduclas2;

     *table->addAttribute("HDUCLAS3",    "photon redistribution matrix")            = "REDIST";

     *table->addAttribute("HDUVERS",    "Version of format (OGIP memo CAL/GEN/92-002a)")            = "1.3.0";

     *table->addAttribute("HDUVERS1",    "Obsolete - included for backwards compatibility")            = "1.0.0";

     *table->addAttribute("HDUVERS1",    "Obsolete - included for backwards compatibility")            = "1.3.0";

     ENDPROC;
}


